#include "../../include/freenet/freenetmasterthread.h"
#include "../../include/option.h"
#include "../../include/uuidgenerator.h"
#include "../../include/stringfunctions.h"
#include "../../include/freenet/unkeyedidcreator.h"
#include "../../include/freenet/identityinserter.h"
#include "../../include/freenet/identityrequester.h"
#include "../../include/freenet/introductionpuzzleinserter.h"
#include "../../include/freenet/identityintroductionrequester.h"
#include "../../include/freenet/introductionpuzzlerequester.h"
#include "../../include/freenet/identityintroductioninserter.h"
#include "../../include/freenet/trustlistinserter.h"
#include "../../include/freenet/trustlistrequester.h"
#include "../../include/freenet/messagelistrequester.h"
#include "../../include/freenet/messagerequester.h"
#include "../../include/freenet/messageinserter.h"
#include "../../include/freenet/messagelistinserter.h"
#include "../../include/freenet/periodicdbmaintenance.h"
#include "../../include/freenet/boardlistinserter.h"
#include "../../include/freenet/boardlistrequester.h"

//#include <zthread/Thread.h>
#include "../../include/pthreadwrapper/thread.h"

#ifdef XMEM
	#include <xmem.h>
#endif

FreenetMasterThread::FreenetMasterThread()
{
	std::string fcpport;

	if(Option::Instance()->Get("FCPHost",m_fcphost)==false)
	{
		m_fcphost="localhost";
		Option::Instance()->Set("FCPHost",m_fcphost);
	}
	if(Option::Instance()->Get("FCPPort",fcpport)==false)
	{
		fcpport="9481";
		Option::Instance()->Set("FCPPort",fcpport);
	}

	// convert fcp port to long, and make sure it's within the valid port range
	if(StringFunctions::Convert(fcpport,m_fcpport)==false)
	{
		m_fcpport=9481;
		Option::Instance()->Set("FCPPort","9481");
	}

	m_receivednodehello=false;

}

FreenetMasterThread::~FreenetMasterThread()
{

}

const bool FreenetMasterThread::FCPConnect()
{
	// we were previosly connected, send FCPDisconnect to objects
	if(m_receivednodehello==true)
	{
		for(std::vector<IFCPConnected *>::iterator i=m_fcpconnected.begin(); i!=m_fcpconnected.end(); i++)
		{
			(*i)->FCPDisconnected();
		}
		m_receivednodehello=false;
	}

	m_log->WriteLog(LogFile::LOGLEVEL_INFO,"FreenetMasterThread::FCPConnect trying to connect to node "+m_fcphost);

	if(m_fcp.Connect(m_fcphost.c_str(),m_fcpport)==true)
	{
		UUIDGenerator uuid;
		std::string clientname="FMSClient-"+uuid.Generate();
		// send ClientHello message to node
		m_fcp.SendMessage("ClientHello",2,"Name",clientname.c_str(),"ExpectedVersion","2.0");

		m_log->WriteLog(LogFile::LOGLEVEL_INFO,"FreenetMasterThread::FCPConnect connected to node");

		return true;
	}
	else
	{
		return false;
	}

}

const bool FreenetMasterThread::HandleMessage(FCPMessage &message)
{
	if(message.GetName()=="NodeHello")
	{
		m_receivednodehello=true;

		// send connected message to all objects, must do this AFTER we received the NodeHello message
		for(std::vector<IFCPConnected *>::iterator i=m_fcpconnected.begin(); i!=m_fcpconnected.end(); i++)
		{
			(*i)->FCPConnected();
		}

		return true;
	}
	if(m_receivednodehello==true)
	{
		bool handled=false;
		std::vector<IFCPMessageHandler *>::iterator i=m_fcpmessagehandlers.begin();
		while(handled==false && i!=m_fcpmessagehandlers.end())
		{
			handled=(*i)->HandleMessage(message);
			i++;
		}

		if(handled==false)
		{
			std::string info("");
			for(std::map<std::string,std::string>::iterator mi=message.begin(); mi!=message.end(); mi++)
			{
				info+="\t\t\t\t"+(*mi).first+"="+(*mi).second+"\r\n";
			}
			m_log->WriteLog(LogFile::LOGLEVEL_DEBUG,"FreenetMasterThread::HandleMessage received unhandled "+message.GetName()+" message.  Message content :\r\n"+info);

			// if unhandled message was alldata - we must retrieve the data
			if(message.GetName()=="AllData")
			{
				long length;
				StringFunctions::Convert(message["DataLength"],length);
				while(m_fcp.Connected() && m_fcp.ReceiveBufferSize()<length)
				{
					m_fcp.Update(1);
				}
				if(m_fcp.Connected())
				{
					char *data=new char[length];
					m_fcp.ReceiveRaw(data,length);
					delete [] data;
				}
			}
		}

		return handled;

	}
	else
	{
		m_log->WriteLog(LogFile::LOGLEVEL_ERROR,"FreenetMasterThread::HandleMessage received "+message.GetName()+" message before NodeHello");
	}

	return false;
}

void FreenetMasterThread::RegisterFCPConnected(IFCPConnected *obj)
{
	m_fcpconnected.push_back(obj);
}

void FreenetMasterThread::RegisterFCPMessageHandler(IFCPMessageHandler *obj)
{
	m_fcpmessagehandlers.push_back(obj);
}

void FreenetMasterThread::RegisterPeriodicProcessor(IPeriodicProcessor *obj)
{
	m_processors.push_back(obj);
}

void FreenetMasterThread::Run()
{

	DateTime lastreceivedmessage;
	DateTime now;
	FCPMessage message;
	bool done=false;

	m_log->WriteLog(LogFile::LOGLEVEL_DEBUG,"FreenetMasterThread::run thread started.");

	Setup();

	do
	{
		if(m_fcp.Connected()==false)
		{
			if(FCPConnect()==false)
			{

				m_log->WriteLog(LogFile::LOGLEVEL_ERROR,"FreenetMasterThread::run could not connect to node.  Waiting 60 seconds.");

				// wait 60 seconds - will then try to connect again
				/*
				try
				{
					ZThread::Thread::sleep(60000);
				}
				catch(...)
				{
					done=true;
				}
				*/
				for(int i=0; i<60 && !IsCancelled(); i++)
				{
					Sleep(1000);
				}
			}
			else
			{
				lastreceivedmessage.SetToGMTime();
			}
		}
		// fcp is connected
		else
		{
			m_fcp.Update(1);

			// check for message on receive buffer and handle it
			if(m_fcp.ReceiveBufferSize()>0)
			{
				message.Reset();
				message=m_fcp.ReceiveMessage();

				if(message.GetName()!="")
				{
					HandleMessage(message);
					lastreceivedmessage.SetToGMTime();
				}
			}

			// let objects do their processing
			for(std::vector<IPeriodicProcessor *>::iterator i=m_processors.begin(); i!=m_processors.end(); i++)
			{
				(*i)->Process();
			}

			// if we haven't received any messages from the node in 10 minutes, something is wrong
			now.SetToGMTime();
			if(lastreceivedmessage<(now-(1.0/144.0)))
			{
				m_log->WriteLog(LogFile::LOGLEVEL_ERROR,"FreenetMasterThread::Run The Freenet node has not responded in 10 minutes.  Trying to reconnect.");
				m_fcp.Disconnect();
			}

		}
//	}while(!ZThread::Thread::interrupted() && done==false);
	}while(!IsCancelled() && done==false);

	m_fcp.Disconnect();

	Shutdown();

	m_log->WriteLog(LogFile::LOGLEVEL_DEBUG,"FreenetMasterThread::run thread exiting.");

}

void FreenetMasterThread::Setup()
{

	// seed random number generator
	srand(time(NULL));

	m_registrables.push_back(new UnkeyedIDCreator(&m_fcp));
	m_registrables.push_back(new IdentityInserter(&m_fcp));
	m_registrables.push_back(new IdentityRequester(&m_fcp));
	m_registrables.push_back(new IntroductionPuzzleInserter(&m_fcp));
	m_registrables.push_back(new IdentityIntroductionRequester(&m_fcp));
	m_registrables.push_back(new IntroductionPuzzleRequester(&m_fcp));
	m_registrables.push_back(new IdentityIntroductionInserter(&m_fcp));
	m_registrables.push_back(new TrustListInserter(&m_fcp));
	m_registrables.push_back(new TrustListRequester(&m_fcp));
	m_registrables.push_back(new MessageListInserter(&m_fcp));
	m_registrables.push_back(new MessageListRequester(&m_fcp));
	m_registrables.push_back(new MessageRequester(&m_fcp));
	m_registrables.push_back(new MessageInserter(&m_fcp));
	m_registrables.push_back(new BoardListInserter(&m_fcp));
	m_registrables.push_back(new BoardListRequester(&m_fcp));
	m_registrables.push_back(new PeriodicDBMaintenance());

	for(std::vector<IFreenetRegistrable *>::iterator i=m_registrables.begin(); i!=m_registrables.end(); i++)
	{
		(*i)->RegisterWithThread(this);
	}

}

void FreenetMasterThread::Shutdown()
{
	// delete each registrable object
	for(std::vector<IFreenetRegistrable *>::iterator i=m_registrables.begin(); i!=m_registrables.end(); i++)
	{
		delete (*i);
	}
}
